/**
 * @file photoshelter.ui.js
 */
(($, Drupal, window) => {
  /**
   * Wrapper object for the current state of the media library.
   */
  Drupal.PhotoshelterUi = {
    /**
     * When a user interacts with the media library we want the selection to
     * persist as long as the media library modal is opened. We temporarily
     * store the selected items while the user filters the media library view or
     * navigates to different tabs.
     */
    currentSelection: [],
  };

  /**
   * Update the photoshelter selection when loaded or media items are selected.
   *
   * @type {Drupal~behavior}
   *
   * @prop {Drupal~behaviorAttach} attach
   *   Attaches behavior to select media items.
   */
  Drupal.behaviors.photoshelterItemSelection = {
    attach(context, settings) {
      const $form = $('.entity-browser-form', context,);
      const currentSelection = Drupal.PhotoshelterUi.currentSelection;

      if (!$form.length) {
        return;
      }

      // Disable submit button since no selection was made yet.
      $('#edit-submit', $form).prop('disabled', true);

      // Enter key submits the search form.
      $("#edit-filters-search").keypress(function (e) {
        if ((e.which && e.which == 13) || (e.keyCode && e.keyCode == 13)) {
          $('#edit-filters-filter').click();
        }
      });

      const $mediaItems = $('.js-photoshelter-item input[type="checkbox"]', $form);

      /**
       * Disable media items.
       *
       * @param {jQuery} $items
       *   A jQuery object representing the media items that should be disabled.
       */
      function disableItems($items) {
        $items
          .prop('disabled', true)
          .closest('.js-photoshelter-item')
          .addClass('photoshelter-item--disabled');
      }

      /**
       * Enable media items.
       *
       * @param {jQuery} $items
       *   A jQuery object representing the media items that should be enabled.
       */
      function enableItems($items) {
        $items
          .prop('disabled', false)
          .closest('.js-photoshelter-item')
          .removeClass('photoshelter-item--disabled');
      }

      /**
       * Update the number of selected items in the button pane.
       *
       * @param {number} remaining
       *   The number of remaining slots.
       */
      function updateSelectionCount(remaining) {
        // When the remaining number of items is a negative number, we allow an
        // unlimited number of items. In that case we don't want to show the
        // number of remaining slots.
        const selectItemsText =
          remaining < 0
            ? Drupal.formatPlural(
                currentSelection.length,
                '1 item selected',
                '@count items selected',
              )
            : Drupal.formatPlural(
                remaining,
                '@selected of @count item selected',
                '@selected of @count items selected',
                {
                  '@selected': currentSelection.length,
                },
              );
        // The selected count div could have been created outside of the
        // context, so we unfortunately can't use context here.
        $('.js-photoshelter-selected-count').html(selectItemsText);
      }

      // Update the selection array and the hidden form field when a media item
      // is selected.
      $(once('photoshelter-item-change', $mediaItems)).on('change', (e) => {
        const id = e.currentTarget.value;

        // Update the selection.
        const position = currentSelection.indexOf(id);
        if (e.currentTarget.checked) {
          // Check if the ID is not already in the selection and add if needed.
          if (position === -1) {
            currentSelection.push(id);
          }
        } else if (position !== -1) {
          // Remove the ID when it is in the current selection.
          currentSelection.splice(position, 1);
        }

        const photoshelterModalSelection = document.querySelector('#photoshelter-selection');

        if (photoshelterModalSelection) {
          // Set the selection in the hidden form element.
          photoshelterModalSelection.value = currentSelection.join();
          $(photoshelterModalSelection).trigger('change');
        }
      });

      // The hidden selection form field changes when the selection is updated.
      $(once('photoshelter-selection-change', $form.find('#photoshelter-selection'))).on('change', (e) => {
        updateSelectionCount(settings.photoshelter.cardinality);

        // Submit button.
        if (currentSelection.length > 0) {
          $('#edit-submit', $form).prop('disabled', false);
        }
        else {
          $('#edit-submit', $form).prop('disabled', true);
        }

        // Prevent users from selecting more items than allowed.
        if (currentSelection.length === settings.photoshelter.cardinality) {
          disableItems($mediaItems.not(':checked'));
          enableItems($mediaItems.filter(':checked'));
        }
        else {
          enableItems($mediaItems);
        }
      });

      // Apply the current selection to the media library view. Changing the
      // checkbox values triggers the change event for the media items. The
      // change event handles updating the hidden selection field for the form.
      currentSelection.forEach((value) => {
        $form
          .find(`input[type="checkbox"][value="${value}"]`)
          .prop('checked', true)
          .trigger('change');
      });

      // Add the selection count to the button pane when a media library dialog
      // is created.
      if (!once('photoshelter-selection-info', 'html').length) {
        return;
      }
      // $(window).on('dialog:aftercreate', () => {
        // Since the dialog HTML is not part of the context, we can't use
        // context here.
        // @todo this doesn't work because dialog.ajax is being called before the modal loads.
        // const $buttonPane = $('.entity-browser-modal .ui-dialog-buttonpane');
        const $buttonPane = $('.entity-browser-form .form-actions');
        if (!$buttonPane.length) {
          return;
        }
        $buttonPane.prepend(Drupal.theme('photoshelterSelectionCount'));
        updateSelectionCount(settings.photoshelter.cardinality);
      // });
    },
  };

  /**
   * Clear the current selection.
   *
   * @type {Drupal~behavior}
   *
   * @prop {Drupal~behaviorAttach} attach
   *   Attaches behavior to clear the selection when the library modal closes.
   */
  Drupal.behaviors.photoshelterModalClearSelection = {
    attach() {
      if (!once('photoshelter-clear-selection', 'html').length) {
        return;
      }
      $(window).on('dialog:afterclose', () => {
        Drupal.MediaLibrary.currentSelection = [];
      });
    },
  };

  /**
   * Theme function for the selection count.
   *
   * @return {string}
   *   The corresponding HTML.
   */
  Drupal.theme.photoshelterSelectionCount = function () {
    return `<div class="photoshelter-selected-count js-photoshelter-selected-count" role="status" aria-live="polite" aria-atomic="true"></div>`;
  };
})(jQuery, Drupal, window);
