<?php

declare(strict_types=1);

namespace Drupal\photoshelter\Controller;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Render\HtmlResponse;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\Core\Url;
use Drupal\photoshelter\Exception\PhotoshelterException;
use Drupal\photoshelter\Photoshelter;
use Drupal\photoshelter\UserErrorTrait;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller for oAuth login.
 */
final class PhotoshelterOAuth extends ControllerBase {

  use UserErrorTrait;

  protected Photoshelter $photoshelter;
  protected LoggerInterface $logger;
  protected ThemeManagerInterface $themeManager;

  public function __construct(
    Photoshelter $photoshelter,
    LoggerChannelFactoryInterface $logger_factory,
    AccountProxyInterface $current_user,
    ThemeManagerInterface $theme_manager
  ) {
    $this->photoshelter = $photoshelter;
    $this->logger = $logger_factory->get('photoshelter');
    $this->currentUser = $current_user;
    $this->themeManager = $theme_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static(
      $container->get('photoshelter'),
      $container->get('logger.factory'),
      $container->get('current_user'),
      $container->get('theme.manager')
    );
  }

  /**
   * The oAuth login controller. A router basically.
   *
   * @return
   */
  public function oAuth(Request $request): Response {
    $homepage_url = Url::fromRoute('<front>', [], ['absolute' => TRUE])->toString(TRUE)->getGeneratedUrl();
    $parameters = $request->query->all();

    $redirect_url_options = [
      'absolute' => TRUE,
    ];
    if (\array_key_exists('destination', $parameters)) {
      $redirect_url_options['query'] = [
        'destination' => $parameters['destination'],
      ];
    }
    $redirect_url = Url::fromRoute('photoshelter.oauth', [], $redirect_url_options)
      ->toString(TRUE)
      ->getGeneratedUrl();

    // The last step first - we got here from the authorization form.
    if (
      \array_key_exists('state', $parameters) &&
      $parameters['state'] === 'ready' &&
      \array_key_exists('code', $parameters)
    ) {
      try {
        $this->photoshelter->oauthToken([
          'code' => $parameters['code'],
          'redirect_uri' => $redirect_url,
        ]);
        $this->messenger()->addStatus($this->t('PhotoShelter authentication completed successfully.'));
      }
      catch (PhotoshelterException $e) {
        $this->userError($e);
      }
    }

    // Otherwise initialize oAuth and display the remote form.
    else {
      try {
        return $this->initiateOauth($homepage_url, $redirect_url);
      }
      catch (PhotoshelterException $e) {
        $this->userError($e);
      }
    }

    if (\array_key_exists('destination', $parameters)) {
      return new RedirectResponse($parameters['destination']);
    }
    return new RedirectResponse($homepage_url);
  }

  private function initiateOauth(string $homepage_url, string $redirect_url): HtmlResponse {
    $response = $this->photoshelter->oauthRegister([
      'name' => 'Drupal',
      'description' => 'The official Drupal - PhotoShelter integration module',
      'redirect_uri' => $redirect_url,
      'email' => $this->currentUser->getEmail(),
      'application_name' => 'Drupal integration',
      'homepage_uri' => $homepage_url,
      'logo_uri' => $homepage_url . $this->themeManager->getActiveTheme()->getLogo(),
    ]);

    $response = $this->photoshelter->oauthAuthorize([
      'api_key' => $this->photoshelter->getApiKey(),
      'client_id' => $response['data']['id'],
      'redirect_uri' => $response['data']['attributes']['redirect_uri'],
      'response_type' => 'code',
      'scope' => 'global',
      'state' => 'ready',
    ]);
    return new HtmlResponse($response['html']);
  }

  public function access(AccountInterface $account): AccessResult {
    if (
      $this->photoshelter->getApiKey() === '' ||
      !$account->hasPermission('use photoshelter')
    ) {
      return AccessResult::forbidden();
    }

    return AccessResult::allowed();
  }

}
