<?php

declare(strict_types=1);

namespace Drupal\photoshelter\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\photoshelter\Exception\PhotoshelterException;
use Drupal\photoshelter\Photoshelter;
use Drupal\photoshelter\UserErrorTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure bynder to enable OAuth based access.
 *
 * @package Drupal\bynder\Form
 */
final class PhotoshelterConfigurationForm extends ConfigFormBase {

  use UserErrorTrait;

  protected Photoshelter $photoshelter;

  public function __construct(
    ConfigFactoryInterface $config_factory,
    Photoshelter $photoshelter
  ) {
    parent::__construct($config_factory);

    $this->photoshelter = $photoshelter;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static(
      $container->get('config.factory'),
      $container->get('photoshelter')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'photoshelter_configuration_form';
  }

  /**
   * {@inheritdoc}
   *
   * @return string[]
   */
  protected function getEditableConfigNames(): array {
    return [Photoshelter::CONFIG_NAME];
  }

  /**
   * {@inheritdoc}
   *
   * @return mixed[]
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $config = $this->config(Photoshelter::CONFIG_NAME);
    $api_key = $config->get('api_key');

    if ($api_key !== NULL && $api_key !== '') {
      try {
        $response = $this->photoshelter->apiVersion();
        $form['health_check'] = [
          '#markup' => $this->t('Connection to PhotoShelter verified, API version: @version', [
            '@version' => $response['version'],
          ]),
        ];

        if (!$this->photoshelter->hasRefreshToken()) {
          $this->messenger()->addWarning($this->t('You are not authenticated in PhotoShelter. @link to be able to sync media.', [
            '@link' => $this->getAuthLink()->toString(),
          ]));
        }
      }
      catch (PhotoshelterException $e) {
        $form['health_check'] = [
          '#markup' => $this->t('Cannot connect to PhotoShelter: @erorr', [
            '@erorr' => $e->getMessage(),
          ]),
        ];
      }
    }

    $form['api_key'] = [
      '#required' => TRUE,
      '#type' => 'textfield',
      '#title' => $this->t('API key'),
      '#default_value' => $api_key,
      '#description' => $this->t('For more information check <a href="@url">PhotoShelter docs</a>.', [
        '@url' => 'https://engineering.photoshelter.com/psapi-v4-doc/#section/Access-and-Authentication/API-key',
      ]),
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $form_state->cleanValues();
    $config = $this->config(Photoshelter::CONFIG_NAME);

    foreach ($form_state->getValues() as $key => $value) {
      $config->set($key, $value);
    }
    $config->save();

    parent::submitForm($form, $form_state);
  }

}
