<?php

declare(strict_types=1);

namespace Drupal\photoshelter\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\photoshelter\Exception\PhotoshelterException;
use Drupal\photoshelter\Photoshelter;
use Drupal\photoshelter\UserErrorTrait;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

final class PhotoshelterUserForm extends FormBase {

  use UserErrorTrait;

  protected Photoshelter $photoshelter;
  protected LoggerInterface $logger;

  public function __construct(
    Photoshelter $photoshelter,
    LoggerChannelFactoryInterface $logger_factory
  ) {
    $this->photoshelter = $photoshelter;
    $this->logger = $logger_factory->get('photoshelter');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static(
      $container->get('photoshelter'),
      $container->get('logger.factory')
    );
  }

  public function getFormId(): string {
    return 'photoshelter_organization_selection';
  }

  /**
   * @return mixed[]
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    if ($this->photoshelter->getApiKey() === '') {
      $form['info'] = [
        '#markup' => $this->t('@link your PhotoShelter API key.', [
          '@link' => Link::createFromRoute($this->t('Please provide'), 'photoshelter.configuration')->toString(),
        ]),
      ];
      return $form;
    }

    // Check if user is logged in to PS first, if not - display info.
    if (!$this->photoshelter->hasRefreshToken()) {
      $form['info'] = [
        '#markup' => $this->t('@link using your PhotoShelter account.', [
          '@link' => Link::createFromRoute($this->t('Please authenticate'), 'photoshelter.oauth', [], [
            'query' => $this->getDestinationArray(),
          ])->toString(),
        ]),
      ];
      return $form;
    }

    // The actual form.
    try {
      $session = $this->photoshelter->userSession();
      $organizations = $this->photoshelter->userOrganization();
    }
    catch (PhotoshelterException $e) {
      $this->userError($e);
      return $form;
    }

    $form['info'] = [
      '#type' => 'table',
      '#header' => [],
      '#rows' => [
        'first' => [
          $this->t('First name'),
          $session['data']['attributes']['first_name'],
        ],
        'last' => [
          $this->t('Last name'),
          $session['data']['attributes']['last_name'],
        ],
        'email' => [
          $this->t('E-mail'),
          $session['data']['attributes']['email'],
        ],
        'zone' => [
          $this->t('Time zone'),
          $session['data']['attributes']['time_zone'],
        ],
      ],
    ];

    $form['actions'] = ['#type' => 'actions'];
    if (\count($organizations) !== 0) {
      $organization_options = [
        '' => $this->t('None'),
      ];
      foreach ($organizations as $organization) {
        $organization_options[$organization['id']] = $organization['name'] . ' (' . $organization['role'] . ')';
      }
      $form['organization'] = [
        '#type' => 'select',
        '#title' => $this->t('Organization'),
        '#options' => $organization_options,
        '#default_value' => $session['data']['attributes']['organization_id'],
      ];

      $form['actions']['update_organization'] = [
        '#type' => 'submit',
        '#value' => $this->t('Change current organization'),
      ];
    }

    $form['actions']['logout'] = [
      '#type' => 'submit',
      '#value' => $this->t('Logout from PhotoShelter'),
    ];

    return $form;
  }

  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $trigger = $form_state->getTriggeringElement();
    $operation = $trigger['#parents'][0];

    try {
      if ($operation === 'update_organization') {
        $organization_id = $form_state->getValue('organization');
        if ($organization_id === '') {
          $this->photoshelter->organizationLogout();
          $this->messenger()->addStatus($this->t('You are now logged out from your current organization.'));
        }
        else {
          $this->photoshelter->organizationAuthenticate([
            'org_id' => $organization_id,
          ]);
          $this->messenger()->addStatus($this->t('You are now logged in to @organization.', [
            '@organization' => $form['organization']['#options'][$organization_id],
          ]));
        }
      }
      elseif ($operation === 'logout') {
        $this->photoshelter->logout();
        $this->messenger()->addStatus($this->t('You have been logged out from PhotoShelter.'));
      }
    }
    catch (PhotoshelterException $e) {
      $this->userError($e);
    }
  }

}
