<?php

declare(strict_types=1);

namespace Drupal\photoshelter;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\photoshelter\Exception\PhotoshelterException;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\ClientException;

final class Photoshelter {

  public const CONFIG_NAME = 'photoshelter.settings';
  private const API_URL = 'https://www.photoshelter.com/psapi';

  /**
   * The number of seconds a token should be refreshed before its expiration
   * time.
   */
  private const SAFETY_MARGIN = 5;

  private const METHODS = [
    'apiVersion' => [
      'path' => 'version',
      'token_required' => FALSE,
    ],
    'oauthRegister' => [
      'method' => 'POST',
      'path' => 'oauth/register',
      'token_required' => FALSE,
    ],
    'oauthAuthorize' => [
      'path' => 'oauth/authorize',
      'token_required' => FALSE,
    ],
    'oauthToken' => [
      'method' => 'POST',
      'path' => 'oauth/token',
      'token_required' => FALSE,
    ],
    'logout' => ['path' => 'authenticate/logout'],
    'userSession' => ['path' => 'user/session'],
    'userOrganization' => ['path' => 'user/organization'],
    'organizationAuthenticate' => [
      'method' => 'POST',
      'path' => 'organization/authenticate',
    ],
    'organizationLogout' => [
      'method' => 'POST',
      'path' => 'organization/logout',
    ],
    'library' => ['path' => 'library'],
    'collection' => ['path' => 'collections/[collection_id]'],
    'collectionChildren' => ['path' => 'collections/[collection_id]/children'],
    'gallery' => ['path' => 'galleries/[gallery_id]'],
    'galleryChildren' => ['path' => 'galleries/[gallery_id]/children'],
    'media' => ['path' => 'media'],
    'mediaDownload' => ['path' => 'media/[media_id]/download'],
    'search' => ['path' => 'search'],
    'metadataFields' => ['path' => 'metadata/fields'],
  ];

  protected ClientInterface $client;
  protected PrivateTempStore $tempstore;
  protected AccountProxyInterface $currentUser;

  private int $requestTime;
  private string $apiKey;
  private string $apiVersion;
  private string $accessToken;
  private int $expireTs;

  public function __construct(
    ClientInterface $client,
    TimeInterface $time,
    ConfigFactoryInterface $config_factory,
    PrivateTempStoreFactory $temp_store_factory,
    AccountProxyInterface $current_user
  ) {
    $this->client = $client;
    $this->tempstore = $temp_store_factory->get('photoshelter');
    $this->currentUser = $current_user;

    $this->requestTime = $time->getRequestTime();

    $config = $config_factory->get(self::CONFIG_NAME);
    $this->apiKey = $config->get('api_key') ?? '';
    $this->apiVersion = $config->get('api_version') ?? 'v4.0';

    $this->accessToken = $this->tempstore->get('access_token') ?? '';
    $this->expireTs = $this->tempstore->get('expire_ts') ?? 0;
  }

  /**
   * Perform an API request.
   *
   * @return mixed[]
   *   Response array.
   */
  private function request($request_options): array {
    if ($this->apiKey === NULL || $this->apiKey === '') {
      throw new PhotoshelterException('PhotoShelter API key is missing.', [], 1);
    }

    if (!$this->currentUser->hasPermission('use photoshelter') && !$this->currentUser->hasPermission('administer photoshelter configuration')) {
      throw new PhotoshelterException('The current user doesn\'t have permissions to use PhotoShelter.', [], 10);
    }

    $options = [
      'headers' => [
        'X-PS-Api-Key' => $this->apiKey,
        'Accept' => 'application/json',
        'content-type' => 'application/x-www-form-urlencoded',
      ],
    ];

    $request_options += [
      'query' => NULL,
      'body' => NULL,
      'token_required' => TRUE,
    ];

    if ($request_options['token_required']) {
      if ($this->accessToken === '') {
        throw new PhotoshelterException('Oauth token required.', [], 2);
      }
      // Check if the token is not expired, if it is, try to get a new one.
      // Failure will throw an exception.
      else {
        if ($this->expireTs < $this->requestTime + self::SAFETY_MARGIN) {
          $this->oauthToken();
        }
      }
      $options['headers']['Authorization'] = 'Bearer ' . $this->accessToken;
    }

    if ($request_options['query'] !== NULL) {
      $options['query'] = $request_options['query'];
    }
    if ($request_options['body'] !== NULL) {
      $options['body'] = \http_build_query($request_options['body']);
    }

    $matches = [];
    if (\preg_match_all('/\[(.+?)\]/', $request_options['path'], $matches)) {
      $replace_pairs = [];
      foreach ($matches[1] as $i => $param) {
        $value = NULL;
        foreach (['query', 'body'] as $parameter_bag) {
          if (
            !\array_key_exists($parameter_bag, $request_options) ||
            $request_options[$parameter_bag] === NULL ||
            !\array_key_exists($param, $request_options[$parameter_bag])
          ) {
            continue;
          }
          $value = $request_options['query'][$param];
          break;
        }
        if ($value === NULL) {
          throw new PhotoshelterException(\sprintf('Path parameter %s not provided.', $param), [], 7);
        }
        $replace_pairs[$matches[0][$i]] = $request_options[$parameter_bag][$param];
      }
      $request_options['path'] = \strtr($request_options['path'], $replace_pairs);
    }

    $uri = \implode('/', [
      self::API_URL,
      $this->apiVersion,
      $request_options['path'],
    ]);

    try {
      $response = $this->client->request($request_options['method'], $uri, $options);

      if ($response->getStatusCode() === 200) {
        $contents = $response->getBody()->getContents();
        $content_type = $response->getHeader('content-type');
        if (\strpos($content_type[0], 'application/json') !== FALSE) {
          $output = \json_decode($contents, TRUE);
          if ($output === NULL) {
            $error_message = \json_last_error_msg();
            if ($error_message === 'No error') {
              return [];
            }
            throw new PhotoshelterException(\sprintf('JSON decode error: %s', $error_message), $request_options);
          }
          return $output;
        }
        elseif (\strpos($content_type[0], 'text/html') !== FALSE) {
          return [
            'html' => $contents,
          ];
        }
        else {
          return [
            'data' => $contents,
          ];
        }
      }
    }
    catch (ClientException $e) {
      $result = \json_decode($e->getResponse()->getBody()->getContents(), TRUE);
      $message_parts = [];
      foreach ($result['errors'] as $error) {
        $detail = '';
        if (\array_key_exists('detail', $error)) {
          $detail = ': ' . $error['detail'];
        }
        $message_parts[] = $error['title'] . $detail . ' (id: ' . $error['id'] . ')';
      }
      // Unauthorized.
      if ($e->getCode() === 401) {
        $code = 6;
        $this->cleanTempstore();
      }
      // API call limit.
      elseif ($e->getCode() === 429) {
        $request_options['retry_after'] = $response->getHeader('Retry-After');
        $code = 9;
      }
      else {
        $code = 0;
      }
      throw new PhotoshelterException(\implode(\PHP_EOL, $message_parts), $request_options, $code);
    }

  }

  private function getUserId(): string {
    $user_id = $this->tempstore->get('user_id');
    if ($user_id === NULL || $user_id === '') {
      $this->userSession();
    }
    return $this->tempstore->get('user_id');
  }

  private function cleanTempstore(): void {
    $this->tempstore->delete('access_token');
    $this->tempstore->delete('refresh_token');
    $this->tempstore->delete('expire_ts');
    $this->tempstore->delete('client_id');
    $this->tempstore->delete('client_secret');
    $this->tempstore->delete('org_id');
  }

  /**
   * @return mixed[]
   *   As from the request() method.
   */
  public function __call(string $method, array $arguments): array {
    if (!\array_key_exists($method, self::METHODS)) {
      throw new PhotoshelterException(\sprintf('Unsupported method %s', $method), [], 3);
    }

    $request_options = self::METHODS[$method];
    if (!\array_key_exists('method', $request_options)) {
      $request_options['method'] = 'GET';
    }

    if (\count($arguments) > 0) {
      $parameters = $arguments[0];

      // Place parameters in the right bag.
      if ($request_options['method'] === 'GET') {
        $key = 'query';
      }
      else {
        $key = 'body';
      }
      $request_options[$key] = [];
      foreach ($parameters as $parameter => $value) {
        $request_options[$key][$parameter] = $value;
      }
    }

    return $this->request($request_options);
  }

  /**
   * Register an oAuth client.
   *
   * @return mixed[]
   */
  public function oauthRegister(array $parameters): array {
    $response = $this->__call('oauthRegister', [$parameters]);
    $this->tempstore->set('client_id', $response['data']['id']);
    $this->tempstore->set('client_secret', $response['data']['attributes']['client_secret']);
    return $response;
  }

  /**
   * Finalize an oAuth process.
   */
  public function oauthToken(array $parameters = []): void {
    $parameters['client_id'] = $this->tempstore->get('client_id');
    $parameters['client_secret'] = $this->tempstore->get('client_secret');

    if ($parameters['client_id'] === NULL || $parameters['client_secret'] === NULL) {
      throw new PhotoshelterException('Trying to request an oAuth token without a registered client.', [], 5);
    }

    if (\array_key_exists('code', $parameters)) {
      $parameters['grant_type'] = 'authorization_code';
    }
    else {
      $refresh_token = $this->tempstore->get('refresh_token');
      if ($refresh_token === NULL) {
        throw new PhotoshelterException('No refresh token available.', [], 4);
      }
      $parameters['grant_type'] = 'refresh_token';
      $parameters['refresh_token'] = $refresh_token;
    }

    $response = $this->__call('oauthToken', [$parameters]);
    $this->tempstore->set('access_token', $response['access_token']);
    $this->tempstore->set('refresh_token', $response['refresh_token']);
    $this->tempstore->set('expire_ts', $response['expires_in'] + $this->requestTime);
  }

  /**
   * We need user_id internally.
   *
   * @return mixed[]
   */
  public function userSession(): array {
    $response = $this->__call('userSession', []);
    $this->tempstore->set('user_id', $response['data']['id']);
    return $response;
  }

  /**
   * We need org_id internally.
   *
   * @return mixed[]
   */
  public function organizationAuthenticate($parameters): array {
    $response = $this->__call('organizationAuthenticate', [$parameters]);
    $this->tempstore->set('org_id', $parameters['org_id']);
    return $response;
  }

  /**
   * We need org_id cleanup.
   *
   * @return mixed[]
   */
  public function organizationLogout(): array {
    $response = $this->__call('organizationLogout', []);
    $this->tempstore->delete('org_id');
    return $response;
  }

  /**
   * Logout from PhotoShelter.
   */
  public function logout(): void {
    $this->__call('logout', []);
    $this->cleanTempstore();
  }

  /**
   * Media search - requires org_id.
   *
   * @return mixed[]
   */
  public function media($parameters): array {
    $org_id = $this->tempstore->get('org_id');
    if ($org_id !== NULL && $org_id !== '') {
      $parameters['org_id'] = $org_id;
    }
    else {
      $parameters['user_id'] = $this->getUserId();
    }
    return $this->__call('media', [$parameters]);
  }

  /**
   * General search - requires org_id.
   *
   * @return mixed[]
   */
  public function search($parameters): array {
    $org_id = $this->tempstore->get('org_id');
    if ($org_id !== NULL && $org_id !== '') {
      $parameters['org_id'] = $org_id;
    }
    else {
      $parameters['user_id'] = $this->getUserId();
    }
    return $this->__call('search', [$parameters]);
  }

  public function isOrganizationAccount(): bool {
    $org_id = $this->tempstore->get('org_id');
    if ($org_id !== NULL && $org_id !== '') {
      return TRUE;
    }
    return FALSE;
  }

  public function hasRefreshToken(): bool {
    $refresh_token = $this->tempstore->get('refresh_token');
    return $refresh_token !== NULL;
  }

  public function getApiKey(): string {
    return $this->apiKey;
  }

}
