<?php

declare(strict_types=1);

namespace Drupal\photoshelter;

use Drupal\Core\Database\Connection;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\Exception\FileWriteException;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\file\FileRepositoryInterface;
use Drupal\photoshelter\Exception\PhotoshelterException;
use Psr\Log\LoggerInterface;

final class PhotoshelterHelper {

  use StringTranslationTrait;

  private const METADATA_TYPES = [
    'iptc',
    'xmp',
    'metadata',
    'custom-metadata',
  ];
  private const METADATA_CID = 'photoshelter_metadata';
  private const METADATA_DATES = [
    'publish_date' => TRUE,
    'creation_date' => TRUE,
    'created_at' => TRUE,
    'modified_at' => TRUE,
  ];
  public const TABLE_NAME = 'photoshelter_usage';

  protected Photoshelter $photoshelter;
  protected Connection $connection;
  protected EntityTypeManagerInterface $entityTypeManager;
  protected LoggerInterface $logger;
  protected AccountProxyInterface $currentUser;
  protected FileSystemInterface $fileSystem;
  protected FileRepositoryInterface $fileRepository;
  protected DateFormatterInterface $dateFormatter;

  private array $metadataTypes;

  public function __construct(
    Photoshelter $photoshelter,
    Connection $connection,
    EntityTypeManagerInterface $entity_type_manager,
    LoggerInterface $logger,
    AccountProxyInterface $current_user,
    FileSystemInterface $file_system,
    FileRepositoryInterface $file_repository,
    DateFormatterInterface $date_formatter
  ) {
    $this->photoshelter = $photoshelter;
    $this->connection = $connection;
    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $logger;
    $this->currentUser = $current_user;
    $this->fileSystem = $file_system;
    $this->fileRepository = $file_repository;
    $this->dateFormatter = $date_formatter;

    // PhotoShelter API throws an error if an individual user requests custom
    // metadata. Due to that behavior we need to store metadata types in a
    // property that depends on user data and remove that type if required.
    $this->metadataTypes = self::METADATA_TYPES;
    if (!$this->photoshelter->isOrganizationAccount()) {
      \array_pop($this->metadataTypes);
    }
  }

  /**
   * @return mixed[]
   *   Drupal db schema for this service.
   */
  public static function schema(): array {
    return [
      self::TABLE_NAME => [
        'description' => "Stores PhotoShelter - Drupal Media ID relation.",
        'fields' => [
          'photoshelter_id' => [
            'type' => 'varchar',
            'length' => 255,
            'not null' => TRUE,
            'description' => "The PhotoShelter media ID.",
          ],
          'media_id' => [
            'type' => 'int',
            'unsigned' => TRUE,
            'description' => "The Drupal media entity ID.",
          ],
        ],
        'primary key' => ['photoshelter_id'],
        'indexes' => [
          'media_id' => ['media_id'],
        ],
      ],
    ];
  }

  /**
   * @param string[] $photoshelter_ids
   *   PhotoShelter IDs array.
   *
   * @return string[]
   *   Drupal media IDs array.
   */
  public function getMediaId(string $photoshelter_id): string {
    $result = $this->connection
      ->select(self::TABLE_NAME, 'pu')
      ->fields('pu', ['media_id'])
      ->condition('photoshelter_id', $photoshelter_id)
      ->execute()
      ->fetchField();
    return $result === FALSE ? '' : $result;
  }

  /**
   * @param string[] $id_pairs
   *   Associative array of Drupal media IDs where keys are PhotoShelter
   *   IDs.
   */
  public function setMediaId(string $media_id, string $photoshelter_id): void {
    $this->connection
      ->merge(self::TABLE_NAME)
      ->key('photoshelter_id', $photoshelter_id)
      ->fields([
        'photoshelter_id' => $photoshelter_id,
        'media_id' => $media_id,
      ])
      ->execute();
  }

  public function deleteMediaId(string $media_id): void {
    $this->connection
      ->delete(self::TABLE_NAME)
      ->condition('media_id', $media_id)
      ->execute();
  }

  /**
   * The heart of the module.
   *
   * @param string[] $photoshelter_ids
   * @param string[] $type_mapping
   * @param bool $save
   *
   * @return \Drupal\media\MediaInterface[]
   */
  public function syncMedia(array $photoshelter_ids, array $type_mapping, bool $save_new = FALSE): array {
    $response = $this->photoshelter->media([
      'media_ids' => \implode(',', $photoshelter_ids),
      // This is limited in items_per_page setting of PhotoshelterLibrary
      // entity_browser widget so we don't need limits here. Set to max.
      'per_page' => 500,
    ]);

    $to_sync = [];
    foreach ($response['data'] as $media_item) {
      if (!\array_key_exists($media_item['type'], $type_mapping)) {
        continue;
      }

      $to_sync[] = [
        'type' => $media_item['type'],
        'id' => $media_item['id'],
        'file_name' => $media_item['attributes']['file_name'],
      ];
    }

    if (\count($to_sync) === 0) {
      return [];
    }

    $media_storage = $this->entityTypeManager->getStorage('media');
    $media_type_storage = $this->entityTypeManager->getStorage('media_type');
    $media_entities = [];
    foreach ($to_sync as $item) {
      // Check if that media item is not synced already.
      $media = NULL;
      $media_id = $this->getMediaId($item['id']);
      if ($media_id !== '') {
        $media = $media_storage->load($media_id);
      }
      if ($media !== NULL) {
        $media_entities[] = $media;
        continue;
      }

      $media_type_id = $type_mapping[$item['type']];
      $media_type = $media_type_storage->load($media_type_id);
      $media = $media_storage->create([
        'bundle' => $media_type->id(),
      ]);
      $source = $media->getSource();
      $field_name = $source->getSourceFieldDefinition($media_type)->getName();
      $field_item = $media->get($field_name)->appendItem();
      $destination_dir = $field_item->getUploadLocation();
      if (!$this->fileSystem->prepareDirectory($destination_dir, FileSystemInterface::CREATE_DIRECTORY)) {
        throw new PhotoshelterException(\sprintf('Unable to save file, directory %s not writeable.', $destination_dir), [], 8);
      }
      $file_uri = $destination_dir . '/' . $item['file_name'];

      $response = $this->photoshelter->mediaDownload([
        'media_id' => $item['id'],
        'download_filetype' => 'original',
      ]);

      try {
        $file = $this->fileRepository->writeData($response['data'], $file_uri, FileSystemInterface::EXISTS_RENAME);
      }
      catch (FileWriteException $e) {
        throw new PhotoshelterException(\sprintf('Unable to save file, to its destination (%s).', $destination_dir), [], 8);
      }

      $field_item->set('target_id', $file->id());
      if ($save_new) {
        $media->save();
        $this->setMediaId($media->id(), $item['id']);
      }
      else {
        $media->photoshelterId = $item['id'];
      }

      $media_entities[] = $media;
    }

    return $media_entities;
  }

  /**
   * @return mixed[]
   *   Library items data.
   */
  public function getLibrary(int $page, int $per_page, array $media_types = []): array {
    $items = [];

    $response = $this->photoshelter->library([
      'page' => $page,
      'per_page' => $per_page,
      'include' => 'key-image,children',
    ]);

    if (\is_null($response)) {
      return $items;
    }

    foreach ($response['data'] as $response_item) {
      $items[] = $this->processContainerItem($response_item, $media_types);
    }

    return [
      'id' => '',
      'type' => 'root',
      'paging' => $this->getPaging($response['meta']),
      'items' => $items,
    ];
  }

  /**
   * @return mixed[]
   *   Collection items data.
   */
  public function getCollection(string $id, int $page, int $per_page, array $media_types = []): array {
    $parameters = [
      'collection_id' => $id,
    ];
    $response = $this->photoshelter->collection($parameters);

    $output = [
      'id' => $response['data']['id'],
      'name' => $response['data']['attributes']['name'],
      'type' => 'collection',
      'items' => [],
    ];

    // Unfortunately API doesn't support paging  when adding `children`
    // as an include so we have to call that endpoint separately.
    // @todo Fix in the API then optimize here.
    $parameters = [
      'collection_id' => $id,
      'page' => $page,
      'per_page' => $per_page,
      'include' => 'key-image,children',
    ];
    $response = $this->photoshelter->collectionChildren($parameters);
    $output['paging'] = $this->getPaging($response['meta']);

    if (!\array_key_exists('data', $response)) {
      return $output;
    }

    foreach ($response['data'] as $child) {
      $output['items'][] = $this->processContainerItem($child, $media_types);
    }

    return $output;
  }

  /**
   * @return mixed[]
   *   Gallery items data.
   */
  public function getGallery(string $id, int $page, int $per_page, array $media_types = []): array {
    $parameters = [
      'gallery_id' => $id,
    ];
    $response = $this->photoshelter->gallery($parameters);

    $output = [
      'id' => $response['data']['id'],
      'name' => $response['data']['attributes']['name'],
      'type' => 'gallery',
      'items' => [],
    ];

    // Unfortunately API doesn't support paging  when adding `children`
    // as an include so we have to call that endpoint separately.
    // @todo Fix in the API then optimize here.
    $include_parts = \array_merge($this->metadataTypes, [
      'link',
    ]);
    $parameters = [
      'galleries' => $id,
      'page' => $page,
      'per_page' => $per_page,
      'include' => \implode(',', $include_parts),
    ];
    if (\count($media_types) !== 0) {
      $parameters['media_type'] = \implode(',', $media_types);
    }

    $response = $this->photoshelter->media($parameters);
    $output['paging'] = $this->getPaging($response['meta']);

    if (!\array_key_exists('data', $response)) {
      return $output;
    }

    foreach ($response['data'] as $child) {
      $output['items'][] = $this->processMediaItem($child);
    }

    return $output;
  }

  /**
   * @return array<string, \Drupal\Core\StringTranslation\TranslatableMarkup>
   */
  private function getMetadataMapping(): array {

    // @todo Currently there's no way to correlate media metadata with metadata
    // titles returned from the API as the number of media metadata is
    // different (36 ipts from metadata/fields and 38 on each image) and keys
    // are also different (for example country_iso on media corresponds to
    // CountryCode in data from metadata/fields). For now we need to define
    // metadata mapping here instead of fetching it from the API.
    //
    // If this is ever fixed in the API, add an aPI call here, set mapping
    // and use injected cache.data to avoid additional API calls. Or, hopefully
    // metadata titles will be returned in iptc as in other metadata types and
    // this can go away completely.
    $metadata_mapping = [
      // iptc.
      'title' => $this->t('Title'),
      'edit_status' => $this->t('Edit Status'),
      'urgency' => $this->t('Urgency'),
      'category' => $this->t('Categories'),
      'keyword' => $this->t('Keywords'),
      'instructions' => $this->t('Special Instruction'),
      'author' => $this->t('Author'),
      'position' => $this->t("Author's Position"),
      'city' => $this->t('City'),
      'state' => $this->t('State'),
      'country' => $this->t('Country'),
      'country_iso' => $this->t('Country Code'),
      'headline' => $this->t('Headline'),
      'trans_ref' => $this->t('Job Id'),
      'credit' => $this->t('Credit Line'),
      'source' => $this->t('Source'),
      'copyright' => $this->t('Copyright'),
      'caption' => $this->t('Description'),
      'caption_writer' => $this->t('Caption Writer'),
      'file_name' => $this->t('Original File Name'),
      'publish_date' => $this->t('Published Date'),
      'ci_city' => $this->t('Creator City'),
      'ci_country' => $this->t('Creator Country'),
      'ci_address' => $this->t('Creator Address'),
      'ci_postal_code' => $this->t('Creator Postal Code'),
      'ci_region' => $this->t('Creator Region'),
      'ci_email' => $this->t('Creator Work Email'),
      'ci_phone' => $this->t('Creator Work Telephone'),
      'ci_url' => $this->t('Creator Work URL'),
      'genre' => $this->t('Genre'),
      'location' => $this->t('Location'),
      'rights' => $this->t('Usage Terms'),
      'scene' => $this->t('Scene'),
      'subject_code' => $this->t('Subject Code'),
      'web_statement' => $this->t('Web Statement'),
      'creation_date' => $this->t('Date Created'),
      'created_at' => $this->t('Created At'),
      'modified_at' => $this->t('Modified At'),

      // metadata.
      'license_expiration_time' => $this->t('License expiration'),
    ];

    return $metadata_mapping;
  }

  /**
   * @return mixed[]
   *   Search results data.
   */
  public function search(string $phrase, int $page, int $per_page, array $media_types = []): array {
    $include_parts = \array_merge($this->metadataTypes, [
      'key-image',
      'link',
    ]);
    $parameters = [
      'terms' => $phrase,
      'include' => \implode(',', $include_parts),
      'page' => $page,
      'per_page' => $per_page,
    ];
    if (\count($media_types) !== 0) {
      $media_types[] = 'collection';
      $media_types[] = 'gallery';
      $parameters['media_type'] = \implode(',', $media_types);
    }

    $response = $this->photoshelter->search($parameters);
    $output = [
      'paging' => $this->getPaging($response['meta']),
      'items' => [],
    ];
    foreach ($response['data'] as $item) {
      if ($item['type'] === 'collection' || $item['type'] === 'gallery') {
        $output['items'][] = $this->processContainerItem($item, $media_types);
      }
      else {
        $output['items'][] = $this->processMediaItem($item);
      }
    }

    return $output;
  }

  /**
   * @return mixed[]
   *   Pager parameters.
   */
  private function getPaging(array $meta): array {
    return [
      'page' => $meta['paging']['page'],
      'n_pages' => $meta['paging']['pages'],
      'total' => $meta['paging']['total'],
    ];
  }

  /**
   * @param mixed[]
   * @return mixed[]
   */
  private function processContainerItem(array $item, array $media_types): array {
    if ($item['type'] === 'collections') {
      $item['type'] = 'collection';
    }

    $processed = [
      'id' => $item['id'],
      'name' => $item['attributes']['name'],
      'type' => $item['type'],
    ];

    if ($item['includes']['key-image']['data']['id'] !== '') {
      $processed['image_url'] = $item['includes']['key-image']['data']['attributes']['image_url'];
      $processed['image_url'] .= '/fit=440x440';
    }

    // Get children data.
    $processed['children'] = $this->getChildrenData($item, $media_types);
    return $processed;
  }

  /**
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup[]
   */
  private function getChildrenData(array $item, array $media_types): string {
    $children_data = [];

    if (\array_key_exists('children', $item['includes']) && \is_array($item['includes']['children']['data'])) {
      $filter_types = $item['type'] === 'gallery' && \count($media_types) !== 0;
      foreach ($item['includes']['children']['data'] as $child) {
        // @todo API can return "collection" or "collections" type for a
        // collection. Correct if ever unified in the API.
        if ($child['type'] === 'collections') {
          $child['type'] === 'collection';
        }

        if ($filter_types && !\in_array($child['type'], $media_types, TRUE)) {
          continue;
        }

        if (!\array_key_exists($child['type'], $children_data)) {
          $children_data[$child['type']] = 0;
        }
        $children_data[$child['type']]++;
      }
    }

    if (\count($children_data) === 0) {
      return (string) $this->t('No items');
    }

    $counts = [];
    foreach ($children_data as $type => $count) {
      if ($type === 'gallery') {
        $counts[] = (string) $this->formatPlural($count, '@count gallery', '@count galleries', [
          '@count' => $count,
        ]);
      }
      elseif ($type === 'collection') {
        $counts[] = (string) $this->formatPlural($count, '@count collection', '@count collections', [
          '@count' => $count,
        ]);
      }
      elseif ($type === 'image') {
        $counts[] = (string) $this->formatPlural($count, '@count image', '@count images', [
          '@count' => $count,
        ]);
      }
      elseif ($type === 'video') {
        $counts[] = (string) $this->formatPlural($count, '@count video', '@count videos', [
          '@count' => $count,
        ]);
      }
      elseif ($type === 'audio') {
        $counts[] = (string) $this->formatPlural($count, '@count audio', '@count audio items', [
          '@count' => $count,
        ]);
      }
    }

    return \implode(', ', $counts);
  }

  /**
   * @param mixed[]
   * @return mixed[]
   */
  private function processMediaItem(array $item): array {
    if ($item['type'] == 'image') {
      $item['includes']['link']['data']['attributes']['link'] .= '/fit=440x440';
    }

    $processed = [
      'id' => $item['id'],
      'type' => 'media',
      'name' => $item['attributes']['file_name'],
      'media_type' => $item['type'],
      'image_url' => $item['includes']['link']['data']['attributes']['link'],
      'metadata' => [],
    ];

    foreach ($this->metadataTypes as $type) {
      if (!\array_key_exists($type, $item['includes'])) {
        continue;
      }
      $metadata = $item['includes'][$type]['data']['attributes'];

      // For some reason custom-metadata has one more container in
      // attributes: custom_metadata. Unify. @todo Solve in the API.
      if ($type === 'custom-metadata') {
        $metadata = $metadata['custom_metadata'];
      }

      // For some reason xmp has one more container in
      // attributes: values. Unify. @todo Solve in the API.
      elseif ($type === 'xmp') {
        $metadata = $metadata['values'];
      }

      // Unfortunately, iptc metadata and metadata metadata doesn't include
      // titles and mapping using API is also impossible.
      // It has to be treated separately.
      // @see self::getMetadataMapping
      if ($type === 'iptc' || $type === 'metadata') {
        $mapping = $this->getMetadataMapping();
        $converted = [];
        foreach ($metadata as $key => $value) {
          if ($value === NULL) {
            continue;
          }
          if (\array_key_exists($key, $mapping)) {
            $name = $mapping[$key];
          }
          else {
            $name = $key;
          }

          // Convert date if applicable.
          if (\array_key_exists($key, self::METADATA_DATES)) {
            $value = $this->dateFormatter->format(\strtotime($value), 'short');
          }
          $converted[] = [
            'name' => (string) $name,
            'value' => $value,
          ];
        }
        $metadata = $converted;
      }

      $converted = [];

      foreach ($metadata as $metadata_item) {
        // @todo custom metadata has "name" key corresponding to the "desc" in
        // xmp metadata. Should be unified.
        $name = \array_key_exists('desc', $metadata_item) ? $metadata_item['desc'] : $metadata_item['name'];
        $converted[$name] = $metadata_item['value'];
      }
      $metadata = $converted;

      // Filter out empty values.
      $metadata = \array_filter($metadata);
      if (\count($metadata) !== 0) {
        $processed['metadata'][$type] = $metadata;
      }
    }

    return $processed;
  }

}
