<?php

declare(strict_types=1);

namespace Drupal\photoshelter\Plugin\EntityBrowser\Widget;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Routing\RedirectDestinationTrait;
use Drupal\entity_browser\Element\EntityBrowserPagerElement;
use Drupal\entity_browser\WidgetBase;
use Drupal\entity_browser\WidgetValidationManager;
use Drupal\photoshelter\Exception\PhotoshelterException;
use Drupal\photoshelter\photoshelterHelper;
use Drupal\photoshelter\UserErrorTrait;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Uses PhotoShelter API to search, browse and select media.
 *
 * @EntityBrowserWidget(
 *   id = "photoshelter_library",
 *   label = @Translation("PhotoShelter library"),
 *   description = @Translation("Adds a PhotoShelter entity browser widget.")
 * )
 */
class PhotoshelterLibrary extends WidgetBase {

  use UserErrorTrait;
  use RedirectDestinationTrait;

  /**
   * Drupal media source to PhotoShelter media type mapping.
   *
   * File is at the end of the array so if any other media source
   * is found before it'll take precedence as more appropriate source.
   *
   * @todo If we ever want to support custom sources this should be a property.
   */
  private const SOURCE_MAPPING = [
    'audio_file' => ['audio'],
    'image' => ['image'],
    'video_file' => ['video'],
    'file' => ['audio', 'image', 'video'],
  ];

  protected PhotoshelterHelper $photoshelterHelper;
  protected LoggerInterface $logger;

  /**
   * @var \Drupal\media\MediaInterface[]
   */
  protected ?array $selectedEntities = NULL;

  /**
   * @var string[]
   */
  private ?array $availableTypes = NULL;

  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EventDispatcherInterface $event_dispatcher,
    EntityTypeManagerInterface $entity_type_manager,
    WidgetValidationManager $validation_manager,
    PhotoshelterHelper $photoshelter_helper,
    LoggerChannelFactoryInterface $logger_factory
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $event_dispatcher, $entity_type_manager, $validation_manager);

    $this->photoshelterHelper = $photoshelter_helper;
    $this->logger = $logger_factory->get('photoshelter');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('event_dispatcher'),
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.entity_browser.widget_validation'),
      $container->get('photoshelter.helper'),
      $container->get('logger.factory')
    );
  }

  /**
   * {@inheritdoc}
   *
   * @return mixed[]
   */
  public function defaultConfiguration(): array {
    return [
      'items_per_page' => 15,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   *
   * @param mixed[] $form
   *
   * @return mixed[]
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['items_per_page'] = [
      '#type' => 'select',
      '#title' => $this->t('Items per page'),
      '#default_value' => $this->configuration['items_per_page'],
      '#options' => ['10' => 10, '15' => 15, '25' => 25, '50' => 50],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function prepareEntities(array $form, FormStateInterface $form_state): array {
    $trigger = $form_state->getTriggeringElement();
    if ($trigger['#parents'][0] === 'filter') {
      return [];
    }

    // Use object property storage to avoid creating entities on validation
    // and submit.
    if ($this->selectedEntities !== NULL) {
      return $this->selectedEntities;
    }

    // As we want to be able to apply nice theming to the elements and have
    // checkboxes on them and renderables included in templates, we need to
    // rely on user input here.
    // @todo Is there a better theming method to have the selection available
    // in $form_state->getValues()?
    $input = $form_state->getUserInput();
    if (!\array_key_exists('items', $input) || !\is_array($input['items']) || \count($input['items']) === 0) {
      return [];
    }

    if (!$this->setTypes($form_state)) {
      return [];
    }

    $this->selectedEntities = $this->photoshelterHelper->syncMedia($input['items'], $this->availableTypes, FALSE);

    return $this->selectedEntities;
  }

  /**
   * @return mixed[]
   */
  public function getForm(array &$original_form, FormStateInterface $form_state, array $additional_widget_parameters): array {
    $form = [];

    if (!$this->setTypes($form_state)) {
      $form['info'] = [
        '#markup' => $this->t('This widget can only work on media entity types that have image, video or audio source. Please check media types  configuration for types that can be referenced here.'),
      ];
      return $form;
    }

    $form = parent::getForm($original_form, $form_state, $additional_widget_parameters);

    $form['#attached']['library'][] = 'photoshelter/widget';

    $context = $form_state->get(['entity_browser', 'widget_context']);
    if (\array_key_exists('cardinality', $context)) {
      $form['#attached']['drupalSettings']['photoshelter']['cardinality'] = $context['cardinality'];
    }

    $form['filters'] = [
      '#type' => 'container',
      '#attributes' => [
        'id' => 'photoshelter-filters-wrapper',
      ],
      '#tree' => TRUE,
    ];

    $form['filters']['parent_element_container'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => [
          'hidden',
        ],
      ],
    ];

    $form['filters']['parent_element_container']['parent_element'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Parent ID'),
      '#attributes' => [
        'class' => [
          'photoshelter-parent-filter',
        ],
      ],
    ];

    $form['filters']['search'] = [
      '#type' => 'search',
      '#title' => $this->t('Search'),
      '#size' => 30,
    ];

    $form['filters']['filter'] = [
      '#type' => 'button',
      '#value' => $this->t('Search'),
      // Limit validation to the filter elements only.
      '#limit_validation_errors' => [['filters']],
      '#attributes' => [
        'class' => [
          'photoshelter-filter-button',
        ],
      ],
    ];

    $modify_breadcrumbs = FALSE;

    // Prepare API call parameters.
    // Pager navigation doesn't include filter values so we need to rely
    // on user input.
    $filters = [];
    $input = $form_state->getUserInput();
    if (\is_array($input) && \array_key_exists('filters', $input)) {
      $filters = $input['filters'];
    }

    if (\array_key_exists('search', $filters) && $filters['search'] !== '') {
      $modify_breadcrumbs = TRUE;
      unset($filters['parent_element_container']['parent_element']);
    }
    elseif (\array_key_exists('parent_element_container', $filters)) {
      if ($filters['parent_element_container']['parent_element'] !== '') {
        $parts = \explode(';', $filters['parent_element_container']['parent_element']);
        $modify_breadcrumbs = TRUE;
        if (\count($parts) >= 2 && $parts[0] !== '') {
          $filters['id'] = $parts[0];
          $filters['type'] = $parts[1];
          unset($filters['search']);
        }
      }
      unset($filters['parent_element_container']['parent_element']);
      $form_state->setValue(['filters', 'search'], '');
    }
    if (\array_key_exists('search', $filters) && $filters['search'] === '') {
      unset($filters['search']);
    }
    unset($filters['filter']);

    // Reset pager on any navigation.
    $trigger = $form_state->getTriggeringElement();
    if ($trigger !== NULL && $trigger['#parents'][0] === 'filters') {
      EntityBrowserPagerElement::setCurrentPage($form_state);
    }

    try {
      $container_data = $this->getData($filters + [
        'media_types' => \array_keys($this->availableTypes),
        'per_page' => (int) $this->configuration['items_per_page'],
        'page' => EntityBrowserPagerElement::getCurrentPage($form_state),
      ]);
    }
    catch (PhotoshelterException $e) {
      $this->userError($e, '_blank');
      return [];
    }

    $form_state->set('photoshelter_items', $container_data['items']);

    $breadcrumb_data = $form_state->get('photoshelter_breadcrumb');
    $breadcrumb_root = [
      'id' => '',
      'type' => 'root',
      'name' => $this->t('Library'),
    ];
    if ($breadcrumb_data === NULL) {
      $modify_breadcrumbs = TRUE;
      $breadcrumb_data = [$breadcrumb_root];
    }

    if ($modify_breadcrumbs) {
      if (\array_key_exists('id', $filters)) {
        $found = FALSE;
        foreach ($breadcrumb_data as $index => $item) {
          if ($item['id'] === $container_data['id']) {
            $found = TRUE;
            $breadcrumb_data = \array_slice($breadcrumb_data, 0, $index + 1);
            break;
          }
        }
        if (!$found) {
          $breadcrumb_data[] = [
            'id' => $container_data['id'],
            'type' => $container_data['type'],
            'name' => $container_data['name'],
          ];
        }
      }
      else {
        $breadcrumb_data = [$breadcrumb_root];
      }
      $form_state->set('photoshelter_breadcrumb', $breadcrumb_data);
    }

    $form['breadcrumbs'] = [
      '#theme' => 'photoshelter_breadcrumbs',
      '#attributes' => [
        'id' => 'photoshelter-breadcrumbs',
        'class' => [
          'breadcrumb__list',
        ],
      ],
      '#title' => $this->t('PhotoShelter breadcrumbs'),
      '#items' => [],
    ];
    foreach ($breadcrumb_data as $breadcrumb_item) {
      $current = \array_key_exists('id', $container_data) && $breadcrumb_item['id'] === $container_data['id'];
      $form['breadcrumbs']['#items'][] = [
        '#wrapper_attributes' => [
          'data-id' => $breadcrumb_item['id'],
          'data-type' => $breadcrumb_item['type'],
          'class' => $current ? ['current'] : [],
        ],
        '#markup' => $breadcrumb_item['name'],
      ];
    }

    $form['items'] = [
      '#type' => 'container',
      '#attributes' => [
        'id' => 'photoshelter-items-wrapper',
      ],
      '#tree' => TRUE,
    ];

    if (\count($container_data['items']) === 0) {
      if (\count($filters) === 0) {
        $user_tab_link = $this->getUserTabLink()->toString();
        $form['items']['info'] = [
          '#markup' => $this->t('There are no results here. Possibliy you need to&nbsp;@link.', [
            '@link' => $user_tab_link,
          ]),
        ];
      }
      else {
        $form['items']['info'] = [
          '#markup' => $this->t('There are no results here.'),
        ];
      }
      return $form;
    }

    foreach ($container_data['items'] as $item) {
      $form['items'][$item['id']] = [
        '#theme' => 'photoshelter_' . $item['type'],
        '#item' => $item,
      ];

      if ($item['type'] === 'media') {
        $form['items'][$item['id']]['#checkbox'] = [
          '#type' => 'checkbox',
          '#name' => 'items[' . $item['id'] . ']',
          '#return_value' => $item['id'],
          '#title' => $item['id'],
          '#title_display' => FALSE,
          '#attributes' => [
            'class' => [
              $context['cardinality'] === 1 ? 'hidden' : ''
            ]
          ]
        ];
      }
    }

    $form['items_selection'] = [
      '#type' => 'hidden',
      '#attributes' => [
        // This is used to identify the hidden field in the form via JavaScript.
        'id' => 'photoshelter-selection',
      ],
    ];

    $form['throbber'] = [
      '#type' => 'markup',
      '#markup' => '<div class="ajax-progress--throbber"><div class="ajax-progress__throbber"></div></div>',
    ];

    if ($container_data['paging']['n_pages'] > 1) {
      $form['pager_eb'] = [
        '#type' => 'entity_browser_pager',
        '#total_pages' => $container_data['paging']['n_pages'],
      ];
    }

    return $form;
  }

  /**
   * Sets PhotoShelter - Drupal media types mapping.
   *
   * @param string[] $context
   */
  private function setTypes(FormStateInterface $form_state): bool {
    $context = $form_state->get(['entity_browser', 'widget_context']);

    if (
      !\is_array($context) ||
      !\array_key_exists('target_entity_type', $context) ||
      $context['target_entity_type'] !== 'media'||
      !\array_key_exists('target_bundles', $context)
    ) {
      return FALSE;
    }

    if ($this->availableTypes !== NULL) {
      return \count($this->availableTypes) !== 0;
    }

    $this->availableTypes = [];
    $media_types = $this->entityTypeManager->getStorage('media_type')->loadMultiple($context['target_bundles']);
    $type_sources = [];
    foreach ($media_types as $media_type) {
      $source_id = $media_type->getSource()->getPluginId();
      $type_sources[$source_id] = $media_type->id();
    }

    foreach (self::SOURCE_MAPPING as $source_id => $photoshelter_types) {
      if (!\array_key_exists($source_id, $type_sources)) {
        continue;
      }
      foreach ($photoshelter_types as $photoshelter_type) {
        // Don't allow overrides. First match lands so if for example image
        // media type will be assigned to image PhotoShelter type, it'll not
        // get overridden by file when it's available.
        if (\array_key_exists($photoshelter_type, $this->availableTypes)) {
          continue;
        }
        $this->availableTypes[$photoshelter_type] = $type_sources[$source_id];
      }
    }

    return \count($this->availableTypes) !== 0;
  }

  /**
   * Gets data for form display.
   *
   * @param string[] $parameters
   *
   * @return mixed[]
   */
  private function getData(array $parameters): array {
    $container_data = [];
    if (\array_key_exists('search', $parameters)) {
      $container_data = $this->photoshelterHelper->search($parameters['search'], $parameters['page'], $parameters['per_page'], $parameters['media_types']);
    }
    elseif (!\array_key_exists('id', $parameters) || $parameters['id'] === '') {
      $container_data = $this->photoshelterHelper->getLibrary($parameters['page'], $parameters['per_page'], $parameters['media_types']);
    }
    elseif ($parameters['type'] === 'collection') {
      $container_data = $this->photoshelterHelper->getCollection($parameters['id'], $parameters['page'], $parameters['per_page'], $parameters['media_types']);
    }
    elseif ($parameters['type'] === 'gallery') {
      $container_data = $this->photoshelterHelper->getGallery($parameters['id'], $parameters['page'], $parameters['per_page'], $parameters['media_types']);
    }

    return $container_data;
  }

  /**
   * {@inheritdoc}
   */
  public function submit(array &$element, array &$form, FormStateInterface $form_state): void {
    $media = $this->prepareEntities($form, $form_state);
    foreach ($media as $media_item) {
      if ($media_item->id() !== NULL) {
        continue;
      }
      $ps_id = $media_item->photoshelterId;
      unset($media_item->photoshelterId);
      $media_item->save();
      $this->photoshelterHelper->setMediaId($media_item->id(), $ps_id);
    }
    $this->selectEntities($media, $form_state);
  }

}
